/*  readact.c  */
/*  (c)  Jure Skvarc  April 1999  */
/* A program which reads stars from the ACT catalog, which are  */
/*    contained in image with center coordinates */
/*  (ra, dec), rotation angle fi  and dimensions (width, height)     */
/*  and outputs their pixel coordinates          */
/*  TAN projection is assumed                    */
#define _POSIX_C_SOURCE  2
#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "../consts.h"
#include "../structs.h"
#include "../functs.h"

extern STATE state;
extern MATCH_OPTIONS match_options;
extern BLINK_FRAME *frame;
/*  Read stars from file name which have declination between declow and dechigh  */
int
read_act_region(char *name, char *accname, double declow, double dechigh, ACT_STAR **stars, int *n, char **error)

{
  FILE *accfp, *fp;
  ACT_ACCEL acc[19];
  int i, k, nacc;
  int lok = 0, hok = 0;
  int indlow = 0, indhigh;

  error = myalloc(1000, "read_act_region", "error");
  /*  Read the accelerator file  */
  accfp = fopen(accname, "r");
  if (accfp == NULL) {
    sprintf(*error, "Can't open the ACT accelerator file %s!", accname);
    return 1;
  }
  nacc = 0;
  while (!feof(accfp)) {
    k = fscanf(accfp, "%f %d %lf", &acc[nacc].limit, &acc[nacc].num, &acc[nacc].reallimit);
    if (k == 3) nacc++;
  }
  fclose(accfp);
  indhigh = acc[nacc - 1].num;
  /*  Find the region which has to be read in  */
  for (i = 1; i < nacc && !(lok && hok); i++) {
    if (!lok && declow >= acc[i - 1].limit && declow < acc[i].limit) {
      indlow = acc[i - 1].num;
      lok = 1;
    }
    if (!hok && dechigh >= acc[i - 1].limit && dechigh < acc[i].limit) {
      indhigh = acc[i].num;
      hok = 1;
    }
  }
  *n = indhigh - indlow;
  /*  Open the catalog file  */
  fp = fopen(name, "r");
  if (fp == NULL) {
    sprintf(*error, "Can't open the ACT catalog file %s!", accname);
    return 1;
  }
  /*  Reserve space for stars  */
  *stars = (ACT_STAR *) malloc(sizeof(ACT_STAR) * *n);
  if (*stars == NULL) {
    sprintf(*error, "Can't reserve space for stars!");
    return 1;
  }
  /*  Read all of the stars between the two indexes  */
  if (fseek(fp, indlow * sizeof(ACT_STAR), SEEK_SET)) {
    sprintf(*error, "Error in fseek for file %s", name);
    return 1;
  }
  k = fread(*stars, sizeof(ACT_STAR), *n, fp);
  if (k != *n) {
    sprintf(*error, "Error in reading star data from file %s!", name);
    return 1;
  }
  fclose(fp);
  return 0;
}

/*  transforms s list of stars in ACT format into a list in "STAR" format */
/*  it copies only those which fall within image limits  */
/*  return 1 if it runs out of space  */
int
extract_act_stars(char *name, char *accname, double declow, double dechigh, STAR *starlist, MAP_IMAGE *image,
		  double jd)

{
  int i, nstar, err, n;
  double xpix, ypix, xinc, yinc;
  ACT_STAR *star;
  int wmin, hmin, wmax, hmax;
  double dt;
  char *error;
  int n1 = state.control[state.control_num];

  nstar = image->nm;
  xinc = image->width / image->w;
  yinc = image->height / image->h;
  
  wmin = -image->extarea * image->w;
  hmin = -image->extarea * image->h;
  wmax =  (1 + image->extarea) * image->w;
  hmax =  (1 + image->extarea) * image->h;
  if (read_act_region(name, accname, declow, dechigh, &star, &n, &error)) {
    if (state.interactive) {
      fl_show_alert(error, "", "", 0);
      free(error);
      return 1;
    }
    else {
      fprintf(stderr, "%s\n", error);
      exit(1);
    }
  }
  for (i = 0; i < n; i++) {
    /*  Check if the star falls into the image  */
    err = 0;
    fits_world_to_pix(star[i].ra, star[i].dec, image->ra, image->dec, 
		0.5 * image->w, 0.5 * image->h, 
		xinc, yinc, image->fi, "-TAN", 
		&xpix, &ypix, &err);
    /*if (!err) printf("%d %f %f %f %f %f %f\n", err, star[i].ra, star[i].dec, star[i].vmag, 
      0, xpix, ypix); */
    if (!err && xpix >= wmin && xpix < wmax && ypix >= hmin && ypix < hmax) {
      float mag;
      switch(frame[n1].match.filter[0]) {
      case 'B':	
      case 'U':	
	mag = star[i].bmag;
	break;
      case 'V':
      case 'W':
      case ' ':	
      case 'R':
      case 'I':
      default:
	mag = star[i].vmag;
	break;
      }
      if (mag > image->minmag && mag < image->maxmag ) {
	dt = (jd - 2451545) / 365.25;
	starlist[nstar].catalog = 'a';
	starlist[nstar].n = starlist[nstar].catnum = star[i].id;
	sprintf(starlist[nstar].catid, "%08.0f", starlist[nstar].catnum);
	starlist[nstar].m = -1;
	/*  Make correction for the proper motion  */
	starlist[nstar].ra = star[i].ra + star[i].pmra * dt / 240.0;
	if (starlist[nstar].ra > 360) starlist[nstar].ra -= 360;
	if (starlist[nstar].ra < 0) starlist[nstar].ra += 360;
	/*  Make correction for the proper motion  */
	starlist[nstar].dec = star[i].dec + star[i].pmdec * dt / 3600.0;
	starlist[nstar].mag = mag;
	starlist[nstar].intensity = 1e9 * pow(10.0, -0.4 * mag);
	starlist[nstar].bluemag = star[i].bmag;
	starlist[nstar].x = xpix - 1;
	starlist[nstar].y = ypix - 1;
	starlist[nstar].q = star[i].q;
	nstar++;
	if (nstar == MAXSTARNUM) {
	  image->nm = nstar;
	  return 1;
	}
      }
    }
  }
  free(star);
  image->nm = nstar;
  return 0;
}


int
get_act_stars(MAP_IMAGE *image, STAR *starlist, char *path, double jd) 

{
  STAR corner[4];
  double x, w;
  double declow, dechigh;
  double ralow[2], rahigh[2];
  int rarange;
  double xpix, ypix;
  int i, j;
  int err;
  char *fname, *accname;
  int indlow, indhigh;
  double xinc = image->width / image->w;
  double yinc = image->height / image->h;

  /*  Reserve space for the file name  */
  if ((fname = malloc(sizeof(char) * (strlen(path) + 50))) == NULL) {
    if (state.interactive) {
      fl_show_alert("Can't reserve space for the file name.", "", "", 0);
      return 1;
    }
    else {
      fprintf(stderr, "Can't reserve space for the file name.\n");
      exit(1);
    }
  }
  if ((accname = malloc(sizeof(char) * (strlen(path) + 50))) == NULL) {
    if (state.interactive) {
      fl_show_alert("Can't reserve space for accelerator file name.", "", "", 0);
      return 1;
    }
    else {
      fprintf(stderr, "Can't reserve space for accelerator file name.\n");
      exit(1);
    }
  }
  calculate_corners(image, corner);
  /*  Find the highest and the lowest declination value  */
  declow = dechigh = corner[0].dec;
  ralow[0] = rahigh[0] = corner[0].ra;
  for (i = 1; i < 4; i++) {
    if (corner[i].dec < declow) declow = corner[i].dec;
    if (corner[i].dec > dechigh) dechigh = corner[i].dec;
    if (corner[i].ra < ralow[0]) ralow[0] = corner[i].ra;
    if (corner[i].ra > rahigh[0]) rahigh[0] = corner[i].ra;
  }
  w = max(image->width, image->height);
  x = declow - image->extarea * w;
  if (x > -90.0) declow = x;
  else declow = -90;
  x = dechigh + image->extarea * w;
  if (x < 90) dechigh = x;
  else dechigh = 90;

  x = ralow[0] - image->extarea * w;
  if (x < 0) {
    rahigh[0] = x + 360.0;
    ralow[0] = rahigh[0] + image->extarea * w;
  }
  else {
    ralow[0] = x;
    x = rahigh[0] + image->extarea * w;
    if (x > 360) {
      rahigh[0] = ralow[0];
      ralow[0] = x - 360.0;
    }
  }

  /* Check if the north pole falls inside the rectangle  */
  err = 0;
  fits_world_to_pix(0.0, 90.0, image->ra, image->dec, 
		    0.5 * image->w, 0.5 * image->h, 
	      xinc, yinc, image->fi, "-TAN", 
	      &xpix, &ypix, &err);
  if (!err && (xpix >= 0.0 && xpix <= image->w && ypix >= 0.0 && ypix <= image->h)) {
      ralow[0] = 0;
      rahigh[0] = 360;
      rarange = 1;
      dechigh = 90.0;
  }
  else {
    /* Check if the south pole falls inside the rectangle  */
    err = 0;
    fits_world_to_pix(0.0, -90.0, image->ra, image->dec,
		      0.5 * image->w, 0.5 * image->h, 
		      xinc, yinc, image->fi, "-TAN", 
		      &xpix, &ypix, &err);
    if (!err && (xpix >= 0.0 && xpix <= image->w && ypix >= 0.0 && ypix <= image->h)) {
      ralow[0] = 0;
      rahigh[0] = 360;
      rarange = 1;
      declow = -90.0;
    }
    else {
      if (rahigh[0] - ralow[0] >= 180) {
	ralow[1] = rahigh[0];
	rahigh[1] = 360;
	rahigh[0] = ralow[0];
	ralow[0] = 0;
	rarange = 2;
      }
      else {
	rarange = 1;
      }
    }
  }
  for (i = 0; i < rarange; i++) {
    /* indexes  of files which contain stars of interest  */
    indlow = floor(ralow[i] / ACT_RAWIDTH);
    indhigh = floor((rahigh[i] - 1e-9) / ACT_RAWIDTH);
    /*  Visit all files which contain stars  */
    for (j = indlow; j <= indhigh; j++) {
      /*  Construct file names  */
      sprintf(fname, "%s/act%02d%02dfts.bin", path, j / 2, (j % 2) * 30);
      sprintf(accname, "%s/act%02d%02d.acc", path, j / 2, (j % 2) * 30);
      extract_act_stars(fname, accname, declow, dechigh, starlist, image, jd);
    }
  }
  free(accname);
  free(fname);
  return 0;
}

