/*  fits2pg                                                    */
/*  Jure Skvarc, 1998                                          */
/*  A program to read header of a fits file and insert the     */
/*  header information into a POSTGRES database                  */
/*=============================================================*/
#ifdef HPUX
#define _INCLUDE_HPUX_SOURCE
#define _INCLUDE_XOPEN_SOURCE
#define _INCLUDE_POSIX_SOURCE
#endif
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <string.h>
#include <getopt.h>
#include <time.h>
#include <unistd.h>

#include <fitsio.h>
#include <libpq-fe.h>
#include "structs.h"
#include "functs.h"

STATE state;

/*  Check if the required keyword was successfully read and exit otherwise  */
void
check_status_strong(char *key, int status)

{
  if (status != 0) {
    fprintf(state.logfile, "Keyword %s is missing. Can't continue!\n", key);
    exit(1);
  }
}

/*  Strip ' from the beginning and end of string, if it exists  */
void
strip(char *string) 

{
  if (string[0] == '\'') {
    strcpy(string, string + 1);
  }
  if (string[strlen(string) - 1] == '\'') {
    string[strlen(string) - 1] = '\0';
  }
}

/*  Divide the input parameter into filename only and the complete path to */
/*  the filename    */
int
resolve_filename(char *fname, char **path, char **name)

{
  char *p, t, *crd;
  
  crd = getcwd(NULL, 0);    /*  Get the current directory  */
  /*  Check if a directory is specified is the file name */
  p = strrchr(fname, '/');
  if (p == 0) {
    /*  There is only a file name  */
    *name = strdup(fname);
  }
  else {
    *name = strdup(p + 1);
    t = p[1];
    p[1] = '\0';
    /*  Change the directory  */
    if (chdir(fname)) {
      fprintf(stderr, "Can't chenge to directory %s\n", fname);
      return 1;
    }
    p[1] = t;
  }
  /*  Get the current directory  */
  *path = getcwd(NULL, 0);
  if (*path == 0) {
      fprintf(stderr, "Can't chenge to directory %s\n", fname);
      return 1;
  }
  /*  Chdir back to the original directory  */
  chdir(crd);
  free(crd);
  return 0;
}


int
main(int argc, char *argv[])
     
{
  int opt;
  FITSFILE f;

  int status = 0;
  int nfound;
  char error[81], comment[81], temp[81];
  int errnum = 0;
  int i, j;
  char a[2000];
  int year, month, day, hour, minute;
  float second;
  time_t t;
  char *insert_date;
  double xc, yc, ac, dc;

  PGconn* pgconn;
  PGresult* pgres;
  int lastid = -1, imageid;
  char *mountpoint = NULL;
  char *archivename = "";
  char *imagepath = NULL;
  char *filename = NULL;
  char *email = NULL;
  char *ucomment = NULL;

  state.server = NULL;
  state.database = "images";
  state.port = NULL;
  state.logfile = stdout;
  state.logname = NULL;
  state.version = "1.0";
  state.verbose = 0;
  /*  Check for command line options  */
  while ((opt = getopt (argc, argv, "l:s:d:p:va:m:e:c:")) != -1) {
    switch(opt) {
    case 'l':
      /*  Name of the log file  */
      state.logname = strdup(optarg);
      if ((state.logfile = fopen(state.logname,"a") )== NULL) {
	fprintf(stderr, "Can not open log file!\n");
	exit(1);
      }
      break;
    case 's':
      /*  Internet address of the server  */
      state.server = strdup(optarg);
      break;
    case 'd':
      /*  Name of the database  */
      state.database = strdup(optarg);
      break;
    case 'p':
      /*  Port number of the server  */
      state.port = strdup(optarg);
      break;
    case 'a':
      /*  archive name  */
      archivename = strdup(optarg);
      break;
    case 'm':
      /*  mount point  */
      mountpoint = strdup(optarg);
      break;
    case 'v':
      /*  Switch on verbose mode  */
      state.verbose = 1;
      break;
    case 'e':
      /*  email address  */
      email = strdup(optarg);
      break;
    case 'c':
      /*  Comment added by a user  */
      ucomment = strdup(optarg);
      break;
    case '?':
      switch(optopt) {
      default:
	fprintf(stderr, "Unknown option -%c ignored!\n", optopt);
	break;
      }
      break;
    }
  }
  if (argc < optind + 1) {
    printf("fits2pg %s     Jure Skvarc, September 1998\n", state.version);
    printf("Usage:: %s [options] fitsfile\n", argv[0]);
    printf("Options:\n");
    printf("-l filename    name of the log file (default stdout)\n");
    printf("-s server      Internet address of the server\n");
    printf("-d database    database name\n");
    printf("-p port        port number\n");
    printf("-a archive     archive name\n");
    printf("-e email       email address\n");
    printf("-c \"comment\"    comment\n");
    printf("-v             verbose mode\n");
    printf("\n");
    exit(1);
  }
  
  /*===================================*/
  /* make a connection to the database */
  /*===================================*/
  pgconn = PQsetdb(state.server, state.port, NULL, NULL, state.database);
  
  /* check to see that the backend connection was successfully made */
  if (PQstatus(pgconn) == CONNECTION_BAD) {
    fprintf(stderr,"Connection to database '%s' failed!", state.database);
    fprintf(stderr,"%s",PQerrorMessage(pgconn));
    PQfinish(pgconn);
    exit(1);
  }
  if (state.verbose) {
    printf("Successfully connected to database %s on server %s\n", 
	   PQdb(pgconn), PQhost(pgconn));
    printf("The port used is %s.\n", PQport(pgconn));
  }
  /*  Find out the highest imageid  */
  /*================================*/
  pgres = PQexec(pgconn, "SELECT * from lastid;");
  if (PQresultStatus(pgres) != PGRES_TUPLES_OK) {
    fprintf(state.logfile, "Error reading id of the last image\n");
    PQclear(pgres);
    PQfinish(pgconn);
    exit(1);
  }
  lastid = strtol(PQgetvalue(pgres, 0, 0), NULL, 10);
  imageid = lastid;
  PQclear(pgres);
  if (state.verbose) {
    printf("Id of the last image: %d\n", lastid);
  }
  for (j = optind; j < argc; j++) {
    f.name = argv[j];
    if (resolve_filename(f.name, &imagepath, &filename)) continue;
    /*  open the input file  */
    /*=======================*/
    fits_open_file(&f.fptr, f.name, 0, &status);
    if (status != 0)  {  /*  check for errors  */
      do  {
	fits_read_errmsg(error);
	fprintf(state.logfile, error);
      } while(strlen(error) > 0);
      return errnum;
    }
    /*  Check if the file is a simple fits file  */
    fits_read_key_log(f.fptr, "SIMPLE", &f.simple, comment, &status);
    check_status_strong("SIMPLE", status);
    if (state.verbose)  {
      printf("SIMPLE=%c\n", f.simple ? 'T' : 'F');
    }
    if (!f.simple) {
      fprintf(state.logfile, "I am not supposed to handle nonsimple files\n");
      return errnum;
    }
    /*  Get depth of a pixel */
    /*=======================*/
    fits_read_key_lng(f.fptr, "BITPIX", &f.bitpix, comment, &status);
    check_status_strong("BITPIX", status);
    if (state.verbose)  {
      printf("BITPIX=%ld\n", f.bitpix);
    }
    /*  Read number of axis  */
    /*=======================*/
    fits_read_key_lng(f.fptr, "NAXIS", &f.naxis, comment, &status);
    check_status_strong("NAXIS", status);
    if (state.verbose)  {
      printf("NAXIS=%ld\n", f.naxis);
    }
    if (f.naxis != 2)  {
      fprintf(state.logfile, "At the moment, I can handle only files with 2 axis\n");
      sprintf(temp, "File %s has %ld axis\n", f.name, f.naxis);
      fprintf(state.logfile, temp);
      do  {
	fits_read_errmsg(error);
	fprintf(state.logfile, error);
      } while(strlen(error) > 0);
      return errnum;
    }
    /*  Read all dimensions at once  */
    /*===============================*/
    fits_read_keys_lng(f.fptr, "NAXIS", 1, f.naxis, f.nax, &nfound, &status);
    check_status_strong("NAXIS", status);
    if (state.verbose)  {
      for (i = 0; i < f.naxis; i++)  {
	printf("NAXIS%d=%ld\n", i + 1, f.nax[i]);
      }
    }


    fits_read_key_dbl(f.fptr, "BZERO", &f.bzero, comment, &status);
    if (status) {
      f.bzero = 0;
      status = 0;
    }
    if (state.verbose) {
      printf("BZERO=%f\n", f.bzero);
    }
    
    fits_read_key_dbl(f.fptr, "BSCALE", &f.bscale, comment, &status);
    if (status) {
      f.bscale = 1;
      status = 0;
    }
    if (state.verbose) {
      printf("BSCALE=%f\n", f.bscale);
    }
    
    /*  Read WCS values  */
    /*===================*/
    fits_read_img_coord(f.fptr,  &f.crval1, &f.crval2, 
			&f.crpix1, &f.crpix2, &f.cdelt1, &f.cdelt2,
			&f.crot, f.ctype, &status);
    /*  Check if everything about WCS was read properly  */
    f.wcs = (status == 0);
    if (!f.wcs) {
      f.crval1 = 0;
      f.crval2 = 0;
      f.crpix1 = 0;
      f.crpix2 = 0;
      f.cdelt1 = 0;
      f.cdelt2 = 0;
      f.crot = 0;
      f.ctype1[0] = '\0';
      f.ctype2[0] = '\0';
      strcpy(f.ctype, "none");
      status = 0;
    }
    else {
      /*  Read CTYPE1 and CTYPE2 keywords  */
      /*===================================*/
      fits_read_keyword(f.fptr, "CTYPE1", f.ctype1, comment, &status);
      fits_read_keyword(f.fptr, "CTYPE2", f.ctype2, comment, &status);
      status = 0;
      strcpy(f.ctype, "-TAN");
    }
    if (state.verbose) {
      printf("CTYPE1=%s\n", f.ctype1);
      printf("CRVAL1=%f\n", f.crval1);
      printf("CRPIX1=%f\n", f.crpix1);
      printf("CDELT1=%f\n", f.cdelt1);
      printf("CTYPE2=%s\n", f.ctype2);
      printf("CRVAL2=%f\n", f.crval2);
      printf("CRPIX2=%f\n", f.crpix2);
      printf("CDELT2=%f\n", f.cdelt2);
    }

    ac = -1;
    dc = 0;
    if (f.wcs) {
      xc = 0.5 * f.nax[0];
      yc = 0.5 * f.nax[1];
      /*  Calculate the center coordinates into ac, dc */
      fits_pix_to_world(xc + 1, yc + 1, f.crval1, f.crval2, 
			f.crpix1, f.crpix2, f.cdelt1, f.cdelt2, 
			f.crot, f.ctype,
			&ac, &dc, &status);
      if (status) {
	ac = -1;
	dc = 0;
	status = 0;
      }
    }



    /*  Read exposure time  */
    /*======================*/
    f.exposure = 0;
    fits_read_key(f.fptr, TFLOAT, "EXPOSURE", &f.exposure, comment, &status);
    if (status != 0) {
      /*  Try again with a nonstandard keyword  */
      status = 0;
      fits_read_key(f.fptr, TFLOAT, "EXPTIME", &f.exposure, comment, &status);
      status = 0;
    }
    if (state.verbose) {
      printf("EXPOSURE=%f\n", f.exposure);
    }
    
    /*  Read time and date   */
    /*=======================*/
    fits_read_keyword(f.fptr, "DATE-OBS", f.date_obs, comment, &status);
    strip(f.date_obs);
    f.is_date_obs = 0;
    if (status != 0) {
      /*  No date  */
      strcpy(f.date_obs, "No DATE-OBS");
      status = 0;
    }
    else 
      f.is_date_obs = 1;
    if (state.verbose) {
      printf("DATE-OBS=%s\n", f.date_obs);
    }
    /*  Try to read different time related keywords  */
    status = 0;
    fits_read_keyword(f.fptr, "UT-START", f.ut_start, comment, &status);
    strip(f.ut_start);
    f.is_ut_start = (status == 0);   /* Flag success on reading start-of-exposure time  */
    if (f.is_ut_start && state.verbose) {
      printf("UT_START=%s\n", f.ut_start);
    }
    status = 0;
    fits_read_keyword(f.fptr, "UT-CENT", f.ut_cent, comment, &status);
    strip(f.ut_cent);
    f.is_ut_cent = (status == 0);   /* Flag success on reading mid-exposure time  */
    if (f.is_ut_cent && state.verbose) {
      printf("UT_CENT=%s\n", f.ut_cent);
    }
    status = 0;
    fits_read_keyword(f.fptr, "UT-END", f.ut_end, comment, &status);
    strip(f.ut_end);
    f.is_ut_end = (status == 0);   /* Flag success on reading end-of-exposure time  */
    if (f.is_ut_end && state.verbose) {
      printf("UT_END=%s\n", f.ut_end);
    }
    status = 0;
    fits_read_keyword(f.fptr, "TIME-BEG", f.time_beg, comment, &status);
    strip(f.time_beg);
    if (status) {
      status = 0;
      fits_read_keyword(f.fptr, "TIME", f.time_beg, comment, &status);
      strip(f.time_beg);
    }
    f.is_time_beg = (status == 0);   /* Flag success on reading start-of-exposure time  */
    if (f.is_time_beg && state.verbose) {
      printf("TIME-BEG=%s\n", f.time_beg);
    }
    status = 0;
    fits_read_keyword(f.fptr, "TIME-END", f.time_end, comment, &status);
    strip(f.time_end);
    f.is_time_end = (status == 0);   /* Flag success on reading start-of-exposure time  */
    if (f.is_time_end && state.verbose) {
      printf("TIME-END=%s\n", f.time_beg);
    }
    /*  Calculate the time of midexposure  */
    calculate_midexposure(&f, &year, &month, &day, &hour, &minute, &second);

    /*  Read the Julian day of observation  */
    f.mjd = 0;
    fits_read_key_dbl(f.fptr, "MJD", &f.mjd, comment, &status);
    if (status) {
      f.mjd = julian_day(year, month, day + hour / 24.0 + minute / 1440.0 +
			 second / 86400.0);
      status = 0;
    }
    if (state.verbose) {
      printf("MJD=%12.6f\n", f.mjd);
    }
    /*==============================*/
    /*  check for a FILTER keyword  */
    /*==============================*/
    fits_read_keyword(f.fptr, "FILTER", f.filter, comment, &status);
    strip(f.filter);
    if (state.verbose) {
      printf("FILTER=%s\n", f.filter);
    }
    status = 0;
    
    /*============================================================*/
    /*  Get observer, instrument and telescope                    */
    /*============================================================*/
    fits_read_keyword(f.fptr, "ORIGIN", f.origin, comment, &status);
    strip(f.origin);
    if (state.verbose) {
      printf("ORIGIN=%s\n", f.origin);
    }
    status = 0;
    fits_read_keyword(f.fptr, "OBSERVER", f.observer, comment, &status);
    strip(f.observer);
    if (state.verbose) {
      printf("OBSERVER=%s\n", f.observer);
    }
    status = 0;
    fits_read_keyword(f.fptr, "TELESCOP", f.telescope, comment, &status);
    strip(f.telescope);
    if (state.verbose) {
      printf("TELESCOP=%s\n", f.telescope);
    }
    status = 0;
    fits_read_keyword(f.fptr, "INSTRUME", f.instrument, comment, &status);
    strip(f.instrument);
    if (state.verbose) {
      printf("INSTRUME=%s\n", f.instrument);
    }
    status = 0;
    fits_read_keyword(f.fptr, "EMAIL", f.email, comment, &status);
    strip(f.email);
    if (status != 0 && email) {
      strcpy(f.email, email);
    }
    if (state.verbose) {
      printf("EMAIL=%s\n", f.email);
    }
    status = 0;
    /*  Close the file  */
    fits_close_file(f.fptr, &status);
    if (status != 0) {
      do  {
	fits_read_errmsg(error);
	fprintf(state.logfile, error);
      } while(strlen(error) > 0);
    }
    if (state.verbose) {
      printf("COMMENT=%s\n", ucomment);
    }
    
    /*======================================================*/
    /*  Everything was properly read from the FITS header,  */
    /*  now put the data into the database                  */
    /*======================================================*/
    /*===================*/
    /*  Prepare a query  */
    /*===================*/
    t = time(NULL);
    insert_date = strdup(asctime(gmtime(&t)) + 4);
    insert_date[strlen(insert_date) - 1] = '\0';
    imageid++;
    sprintf(a, "INSERT INTO images VALUES (%d, '%s', '%s', '%s', '%s', %ld, \
%ld, %ld, %f, %f, '%s', \
%g, %g, %g, %g, %g, %g, %g, \
'%s', %g, %g, %f, \
'%04d-%02d-%02d %02d:%02d:%05.2f GMT', %12.6f, '%s', \
'%s', '%s', '%s', '%s', \
'%s', '%s', '%s')",
	    imageid, mountpoint, archivename, imagepath, filename, f.bitpix,
	    f.nax[0], f.nax[1], f.bzero, f.bscale, f.ctype, 
	    f.crval1, f.crpix1, f.cdelt1, f.crval2, f.crpix2, f.cdelt2, f.crot,
	    "", ac, dc, f.exposure, 
	    year, month, day, hour, minute, second, f.mjd, insert_date, 
	    f.filter, f.origin, f.instrument, f.telescope,
	    f.observer, f.email, (ucomment) ? ucomment : "");
    free(insert_date);
    if (state.verbose) {
      printf("INSERT statement:=%s\n", a);
    }
    /*  Insert the FITS header values into the database  */
    /*==============================================*/
    pgres = PQexec(pgconn, a);
    if (PQresultStatus(pgres) != PGRES_COMMAND_OK) {
      fprintf(stderr,"\nINSERT failed for image %s\n", f.name);
      fprintf(stderr,"%s\n", PQerrorMessage(pgconn));
      imageid--;
      sprintf(a, "UPDATE lastid SET lastid = %d", imageid); 
      pgres = PQexec(pgconn, a);
      if (PQresultStatus(pgres) != PGRES_COMMAND_OK) {
	fprintf(stderr, "Problem with setting of the index!\nExamine the database!\n");
	fprintf(stderr,"%s\n", PQerrorMessage(pgconn));
	fprintf(state.logfile, "Problem with setting of the index!\nExamine the database!\n");
	fprintf(state.logfile,"%s\n", PQerrorMessage(pgconn));
      }
      /*  PQfinish(pgconn);
	  exit(1); */
    }     
    PQclear(pgres);
    free(imagepath);
    free(filename);
  }
  sprintf(a, "UPDATE lastid SET lastid = %d", imageid); 
  pgres = PQexec(pgconn, a);
  if (PQresultStatus(pgres) != PGRES_COMMAND_OK) {
    PQclear(pgres);
    fprintf(stderr, "Problem with setting of the index!\nExamine the database!\n");
fprintf(state.logfile, "Problem with setting of the index!\nExamine the database!\n");
  }
  PQfinish(pgconn);
  if (state.verbose) {
    printf("Connection closed!\n");
  }
  exit(0);
}



