/*  fits.c  for Fitsblink */
#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include <fitsio.h>
#include "functs.h"
#include "consts.h"

extern OBSERVATORY observatory;
extern STATE state;
/* 
    The datatype parameter can have one of the following symbolic constant 
    values: TSTRING, TLOGICAL, TBYTE, TSHORT, TINT, TLONG, TFLOAT, TDOUBLE.
*/
int
read_fits(char *name, PIXEL **image1, int *xmax, int *ymax,  FITSFILE *f, 
	  char **errors)
     
{
  int pixelsize;
  int imagesize;
  
  int status = 0;
  int nfound, anyf;
  char error[FLEN_CARD], comment[FLEN_CARD], temp[FLEN_CARD];
  int verbose = 0;
  int i;
  PIXEL *image;
  int errnum = 0;
  int noofkeys;

  /*  open the file  */
  fits_open_file(&f->fptr, name, READONLY, &status);
  if (status != 0)  {
    do  {
      fits_read_errmsg(error);
      errors[errnum++] = strdup(error);
    } while(strlen(error) > 0 && errnum < ERROR_MAX - 1);
    if (errnum == ERROR_MAX - 1) {
      errors[errnum] = "And there may be other problems, too";
    }
    return errnum;
  }
  /*  Check if the file is simple fits file  */
  fits_read_key_log(f->fptr, "SIMPLE", &f->simple, comment, &status);
  if (verbose)  {
    printf("SIMPLE=%c\n", f->simple ? 'T' : 'F');
  }
  if (!f->simple) {
    errors[errnum] = "I am not supposed to handle nonsimple files";
    return errnum;
  }
  /*  Get depth of a pixel */
  fits_read_key_lng(f->fptr, "BITPIX", &f->bitpix, comment, &status);
  if (verbose)  {
    printf("BITPIX=%ld\n", f->bitpix);
  }
  /*  Read number of axis  */
  fits_read_key_lng(f->fptr, "NAXIS", &f->naxis, comment, &status);
  if (verbose)  {
    printf("NAXIS=%ld\n", f->naxis);
  }
  if (f->naxis != 2)  {
    errors[errnum++] = "At the moment, I can handle only files with 2 axis";
    sprintf(temp, "File %s has %ld axis\n", name, f->naxis);
    errors[errnum++] = strdup(temp);
    do  {
      fits_read_errmsg(error);
      errors[errnum++] = strdup(error);
    } while(strlen(error) > 0 && errnum < ERROR_MAX - 1);
    if (errnum == ERROR_MAX - 1) {
      errors[errnum] = "And there may be other problems, too.";
    }
    return errnum;
  }
  /*  Read dimensions at once  */
  fits_read_keys_lng(f->fptr, "NAXIS", 1, f->naxis, f->nax, &nfound, &status);
  if (verbose)  {
    for (i = 0; i < f->naxis; i++)  {
      printf("NAXIS%d=%ld\n", i + 1, f->nax[i]);
    }
  }
  /*  Size of pixel in bytes  */
  pixelsize = sizeof(PIXEL);
  /*  Determine number of pixels  */
  imagesize = 1;
  for (i = 0; i < f->naxis; i++) imagesize *= f->nax[i];
  if (status != 0) {
    do  {
      fits_read_errmsg(error);
      errors[errnum++] = strdup(error);
    } while(strlen(error) > 0 && errnum < ERROR_MAX - 1);
    if (errnum == ERROR_MAX - 1) {
      errors[errnum] = "And there may be other problems, too.";
    }
  }
  /*============================================================*/
  /*  Try to read WCS coordinates                               */
  /*============================================================*/
  status = 0;
  if (verbose)  {
    printf("Reading WCS\n");
  }
  fits_read_img_coord(f->fptr,  &f->crval1, &f->crval2, 
       &f->crpix1, &f->crpix2, &f->cdelt1, &f->cdelt2,
       &f->crot, f->ctype, &status);
  /*  Check if everything about WCS was read properly  */
  f->wcs = (status == 0);
  if (f->wcs) {
    /*  if ctype is empty, assume -TAN  */
    if (strlen(f->ctype) == 0) {
      strcpy(f->ctype, "-TAN");
    }
    /*  Read CTYPE1 keyword  */
    status = 0;
    fits_read_keyword(f->fptr, "CTYPE1", f->ctype1, comment, &status);
    /*  if CTYPE1 is DEC, than exchange the crval1 and crval2 values  */
    if (strncmp(f->ctype1 + 1, "DEC", 3) == 0) {
      double t;
      t = f->crval1;
      f->crval1 = f->crval2;
      f->crval2 = t;
    }
  }
  status = 0;
  /*  Check if WCS values were already computed using a star match with a catalog */
  f->matched = 0;
  fits_read_key_log(f->fptr, "MATCHED", &f->matched, comment, &status);
  if (verbose)  {
    printf("MATCHED=%c\n", f->matched ? 'T' : 'F');
  }
  /*============================================================*/
  /*  Get date and time of observation and exposure time        */
  /*============================================================*/
  status = 0;
  fits_read_keyword(f->fptr, "DATE-OBS", f->date_obs, comment, &status);
  f->is_date_obs = (status == 0);   /* Flag success on reading date  */
  status = 0;
  fits_read_key(f->fptr, TFLOAT, "EXPOSURE", &(f->exposure), comment, &status);
  f->is_exposure = (status == 0);   /* Flag success on reading exposure time  */
  if (status != 0) {
    /*  Try again with a nonstandard keyword  */
    status = 0;
    fits_read_key(f->fptr, TFLOAT, "EXPTIME", &(f->exposure), comment, &status);
    f->is_exposure = (status == 0);   /* Flag success on reading exposure time  */
  }
  /*  Try to read the camera gain  */
  status = 0;
  fits_read_key(f->fptr, TFLOAT, "GAIN", &(f->gain), comment, &status);
  f->is_gain = (status == 0);   /* Flag success on reading exposure time  */
  /*  Try to read the camera read out noise  */
  status = 0;
  fits_read_key(f->fptr, TFLOAT, "RDNOISE", &(f->rdnoise), comment, &status);
  f->is_rdnoise = (status == 0);   /* Flag success on reading exposure time  */
  status = 0;
  f->ut_start[0] = '\0';
  fits_read_keyword(f->fptr, "UT", f->ut_start, comment, &status);
  f->is_ut_start = (status == 0);   /* Flag success on reading start-of-exposure time  */
  if (!f->is_ut_start) {
      status = 0;
      fits_read_keyword(f->fptr, "UT-START", f->ut_start, comment, &status);
      f->is_ut_start = (status == 0);   /* Flag success on reading start-of-exposure time  */
  }
  status = 0;
  fits_read_keyword(f->fptr, "UT-CENT", f->ut_cent, comment, &status);
  f->is_ut_cent = (status == 0);   /* Flag success on reading mid-exposure time  */
  status = 0;
  fits_read_keyword(f->fptr, "UT-END", f->ut_end, comment, &status);
  f->is_ut_end = (status == 0);   /* Flag success on reading end-of-exposure time  */
  status = 0;
  fits_read_keyword(f->fptr, "TIME-BEG", f->time_beg, comment, &status);
  f->is_time_beg = (status == 0);   /* Flag success on reading start-of-exposure time  */
  status = 0;
  fits_read_keyword(f->fptr, "TIME-END", f->time_end, comment, &status);
  f->is_time_end = (status == 0);   /* Flag success on reading start-of-exposure time  */
  /*============================================================*/
  /*  Get observer, instrument and telescope                    */
  /*  If they are not present, insert values from the settings  */
  /*  file                                                      */
  /*============================================================*/
  status = 0;
  fits_read_keyword(f->fptr, "OBSERVER", f->observer, comment, &status);
  if (status != 0) {
    strcpy(f->observer, observatory.observer);
    status = 0;
  }
  else {
    strcpy(observatory.observer, f->observer);
  }
  fits_read_keyword(f->fptr, "TELESCOP", f->telescope, comment, &status);
  if (status != 0) {
    strcpy(f->telescope, observatory.telescope);
    status = 0;
  }
  else {
    strcpy(observatory.telescope, f->telescope);
  }
  fits_read_keyword(f->fptr, "INSTRUME", f->instrument, comment, &status);
  if (status != 0) {
    strcpy(f->instrument, observatory.instrument);
    status = 0;
  }
  /*=========================================*/
  /*  Get filter name, first letter only     */
  /*=========================================*/
  fits_read_keyword(f->fptr, "FILTER", temp, comment, &status);
  if (status == 0) {
    /*  Search for the first letter */
    i = 0;
    while (temp[i] > 0 && (temp[i] == ' ' || temp[i] == '\'')) i++;
    if (temp[i] != 0) strncpy(f->filter, temp + i, 1);
    else f->filter[0] = '*';
  }
  else {
    status = 0;
    f->filter[0] = '*';
  }
  f->filter[1] = '\0';
  /*===================================================*/
  /*  Read the object name                             */
  /*===================================================*/
  fits_read_keyword(f->fptr, "OBJECT", f->object, comment, &status);
  /*  Erase object  */
  f->object[0] = '\0';
  if (status != 0) {
    f->object[0] = '\0';
    status = 0;
  }
  /*======================================================*/
  /*  Now read the all of the header lines and store them */
  /*======================================================*/
  /* Get number of keys  */
  fits_get_hdrspace(f->fptr, &noofkeys, NULL, &status);
  noofkeys++;  /*  Add one for the END keyword  */
  f->nkeys = noofkeys;
  /*  reserve space for the header  */
  f->key = (char **) myalloc(sizeof(char *) * noofkeys, "read_fits", "f->key");
  /*  Read all header lines  */
  for (i = 0; i < noofkeys; i++) {
    /* Reserve space for a line  */
    f->key[i] = (char *) myalloc(sizeof(char *) * FLEN_CARD, "read_fits", 
				 "f->key[i]");
    fits_read_record(f->fptr, i + 1, f->key[i], &status);
  }
  /*================================================================*/
  /*  Read in the image                                             */
  /*================================================================*/
  status = 0;
  /*  Allocate memory for the image  */
  if (verbose)  {
    printf("Allocating space for the image\n");
  }
  image = (PIXEL *) myalloc(pixelsize * imagesize, "read_fits", "image");
  /*  Transfer the image from disk to memory */
  if (verbose)  {
    printf("Reading the image\n");
  }
  fits_read_2d_sht(f->fptr, 0, 0, f->nax[0], f->nax[0], f->nax[1], image, &anyf, &status); 
  if (verbose)  {
    printf("Image read\n");
  }
  if (status == NUM_OVERFLOW && f->bitpix <= 16) {
    /*  Image reading did not succeed due to numerical overflow  */
    double bscale, bzero = 0;
    status = 0;
    /*  Find out bscale parameter  */
    fits_read_key(f->fptr, TDOUBLE, "BSCALE", &bscale, comment, &status);
    /*  Try again with a BZERO set to 0  */
    fits_set_bscale(f->fptr, bscale, bzero, &status); 
    fits_read_2d_sht(f->fptr, 0, 0, f->nax[0], f->nax[0], f->nax[1], image, &anyf, &status);
    for (i = 0; i < f->nax[0] * f->nax[1]; i++) {
      image[i] -= 32768;
    }
  }
  *xmax = f->nax[0];
  *ymax = f->nax[1];
  *image1 = image;
  /*  Close the file  */
  fits_close_file(f->fptr, &status);
  if (status != 0) {
    do  {
      fits_read_errmsg(error);
      errors[errnum++] = strdup(error);
    } while(strlen(error) > 0 && errnum < ERROR_MAX - 1);
    if (errnum == ERROR_MAX - 1) {
      errors[errnum] = "And there may be other problems, too.";
    }
  }
  if (verbose)  {
    printf("End of read_fits\n");
  }
  return errnum;
}

/*====================================================================*/
/*  Write a FITS file from the data contained in the given structure  */
/*====================================================================*/
int
write_fits(BLINK_FRAME *frame, int n, char **errors)

{
  int status = 0;
  fitsfile *fptr;
  int i, i0, errnum = 0;
  char error[FLEN_CARD];
  long nax1, nax2;
  double crpix1, crpix2;

  fits_create_file(&fptr, frame[n].name, &status);
  if (!status) {
    /* First write the header which is already stored in the frame.fitsfile */
    for (i = 0; i < frame[n].fitsfile.nkeys - 1; i++) {
      fits_write_record(fptr, frame[n].fitsfile.key[i], &status);
      if (status) break;
    }
    nax1 = frame[n].fits.x2 - frame[n].fits.x1 + 3;
    nax2 = frame[n].fits.y2 - frame[n].fits.y1 + 3;
    crpix1 = frame[n].fitsfile.crpix1 - frame[n].fits.x1 + 1;
    crpix2 = frame[n].fitsfile.crpix2 - frame[n].fits.y1 + 1;
    /*  Update keyword value that might have changed  */
    fits_update_key(fptr, TLONG, "NAXIS1", &nax1, 
		    "Updated by fitsblink", &status);
    fits_update_key(fptr, TLONG, "NAXIS2", &nax2, 
		    "Updated by fitsblink", &status);
    fits_update_key(fptr, TDOUBLE, "CRVAL1", &frame[n].fitsfile.crval1, 
		    "Updated by fitsblink", &status);
    fits_update_key(fptr, TDOUBLE, "CRVAL2", &frame[n].fitsfile.crval2, 
		    "Updated by fitsblink", &status);
    fits_update_key(fptr, TDOUBLE, "CRPIX1", &crpix1, 
		    "Updated by fitsblink", &status);
    fits_update_key(fptr, TDOUBLE, "CRPIX2", &crpix2, 
		    "Updated by fitsblink", &status);
    fits_update_key(fptr, TDOUBLE, "CDELT1", &frame[n].fitsfile.cdelt1, 
		    "Updated by fitsblink", &status);
    fits_update_key(fptr, TDOUBLE, "CDELT2", &frame[n].fitsfile.cdelt2, 
		    "Updated by fitsblink", &status);
    fits_update_key(fptr, TDOUBLE, "CROTA1", &frame[n].fitsfile.crot, 
		    "Updated by fitsblink", &status);
    fits_update_key(fptr, TDOUBLE, "CROTA2", &frame[n].fitsfile.crot, 
		    "Updated by fitsblink", &status);
    /*fits_update_key(fptr, TSTRING, "CTYPE1", &frame[n].fitsfile.ctype1, 
      "Updated by fitsblink", &status);*/
    /*  Write the image  */
    i0 = frame[n].fits.y1 - 1;
    for (i = i0 ; i <= frame[n].fits.y2; i++)
      fits_write_img(fptr, TSHORT, 
		     (i - i0) * nax1 + 1/*frame[n].fits.x1*/, 
		     nax1, frame[n].fits.pi[i] + frame[n].fits.x1 - 1, 
		     &status);
    fits_close_file(fptr, &status);
  }
  errnum = 0;
  if (status != 0) {
    do  {
      fits_read_errmsg(error);
      errors[errnum++] = strdup(error);
    } while(strlen(error) > 0 && errnum < ERROR_MAX - 1);
    if (errnum == ERROR_MAX - 1) {
      errors[errnum] = "And there may be other problems, too.";
    }
  }
  return errnum;
}




/*===========================================================================*/
/*  Read a FITS table heasderand make a structure which will hold the result */
/*===========================================================================*/
int
read_table_header(char *name, FITS_TABLE *table, FL_OBJECT *header, 
		  FL_OBJECT *choice, FL_OBJECT *fields, 
		  FL_OBJECT *xaxis, FL_OBJECT *yaxis)

{
  int i;
  long nrows;
  int ncol;
  fitsfile *fptr;
  int hdutype;
  int status = 0, status1;
  char colname[80];
  int colnum, type;
  long repeat, width;
  int noofkeys;

  /*  Open the file  */
  fits_open_file(&fptr, name, READONLY, &status);
  if (status) {
    if (state.interactive) {
      fl_show_alert("Can not open file", name, "for reading or not FITS file!", 0);
      free(name);
      return 1;
    }
    else {
      fprintf(stderr, "Can not open file %s for reading or not FITS file!\n", name);
      free(name);
      return 1;
    }	
  }

  /*  Try to find a table extension  */
  /*  Move to the next HDU extension  */
  i = 1;
  do {
    i++;
    fits_movrel_hdu(fptr, 1, NULL, &status);
    fits_get_hdu_type(fptr, &hdutype, &status);
    /*fl_add_choice(choice, */
  } while(!(hdutype == BINARY_TBL || hdutype == ASCII_TBL || status != 0));
  if (status) {
    if (state.interactive) {
      fl_show_alert("No tables exist in this file!", "", "", 0);
      return 1;
    }
    else {
      fprintf(stderr, "No tables exist in file %s!\n", name);
      return 1;
    }	
  }
  table->hdu = i;
  /*  Get table dimensions  */
  /*  Read the number of columns in table  */
  fits_get_num_rows(fptr, &nrows, &status);
  /*  Read the number of items in the table  */
  fits_get_num_cols(fptr, &ncol, &status);
  if (status) {
    if (state.interactive) {
      fl_show_alert("Could not read table dimensions!", "", "", 0);
      return 1;
    }
    else {
      fprintf(stderr, "Could not read table dimensions in file %s!\n", name);
      return 1;
    }	
  }
  create_table(table, name, ncol, nrows);
  /*  Read column names  */
  status1 = 0;
  if (state.interactive) {
    fl_freeze_form(fields->form);
    fl_clear_browser(fields);
    fl_clear_choice(xaxis);
    fl_clear_choice(yaxis);
  }
  for (i = 0; i < table->ncol; i++) {
    fits_get_colname(fptr, CASEINSEN, "*", colname, &colnum, &status1);
    table->colnames[i] = strdup(colname);
    if (state.interactive) {
      fl_addto_choice(xaxis, colname);
      fl_addto_choice(yaxis, colname);
      fl_addto_browser(fields, colname);
      fl_select_browser_line(fields, i + 1);
    }
    width = 0;
    type = 0;
    status = 0;
    fits_get_coltype(fptr, colnum, &type, &repeat, &width, &status);
    table->type[i] = type;
    table->width[i] = width;
  }
  if (state.interactive) {
    fl_set_browser_topline(fields, 1);
    fl_unfreeze_form(fields->form);
  }

  /*fits_read_col_str(fptr, 1, 1, 1, table.ncol, "", &array, &anyf, &status);*/
  /*  status 237 means multiple match for *, what is not an error  */
  if (status == 237) status = 0;
  fits_report_error(stderr, status);
  
  /* Get number of keys  */
  fits_get_hdrspace(fptr, &noofkeys, NULL, &status);
  noofkeys++;  /*  Add one for the END keyword  */
  table->nkeys = noofkeys;
  if (state.interactive) {
    fl_freeze_form(header->form);
    fl_clear_browser(header);
  }
  /*  reserve space for the header  */
  table->key = (char **) myalloc(sizeof(char *) * noofkeys, "read_header_table", "table->key");
  /*  Read all header lines  */
  for (i = 0; i < noofkeys; i++) {
    /* Reserve space for a line  */
    table->key[i] = (char *) myalloc(sizeof(char *) * FLEN_CARD,
				     "read_header_table", "table->key[i]");
    fits_read_record(fptr, i + 1, table->key[i], &status);
    /* Write it into the header browser  */
    if (state.interactive) {
      fl_addto_browser(header, table->key[i]);
    }
  }
  if (state.interactive) {
    fl_set_browser_topline(header, 1);
    fl_unfreeze_form(header->form);
  }
  fits_close_file(fptr, &status);
  return 0;
}


/*  Read FITS table contents  */
/*  Assumes that the table structure is already filled and selected vector */
/*  is filled  */
int
read_table_contents(FITS_TABLE *table, FL_OBJECT *titles, FL_OBJECT *browser,
		    FL_OBJECT *selected, FL_OBJECT *choice) 

{
  int i, j, j1;
  fitsfile *fptr;
  int hdutype;
  int status = 0;
  void *element;
  int maxwidth;
  int retval = 0;
  char cnline[MAXBROWSERLENGTH + 1];
  int unknown;

  if (state.interactive) {
    fl_clear_browser(browser);
    fl_clear_browser(titles);
  }
  /*  Find data type with the largest required storage */
  maxwidth = 0;
  for (i = 0; i < table->ncol; i++) 
    if (table->width[i] > maxwidth) 
      maxwidth = table->width[i];
  /*  Reserve space  */
  element = myalloc(maxwidth + 1, "read_table_contents", "element");
  /*  Open the file  */
  fits_open_file(&fptr, table->name, READONLY, &status);
  if (status) {
    if (state.interactive) {
      fl_show_alert("Can not open file", table->name, "for reading!", 0);
      retval = 1;
      goto end_rtc;
    }
    else {
      fprintf(stderr, "Can not open file %s for reading!\n", table->name);
      retval = 1;
      goto end_rtc;
    }	
  }

  /*  Move to the HDU extension which contains the table */
  fits_movabs_hdu(fptr, table->hdu, NULL, &status);
  fits_get_hdu_type(fptr, &hdutype, &status);
  if (status || (hdutype != BINARY_TBL && hdutype != ASCII_TBL)) {
    if (state.interactive) {
      fl_show_alert("No tables exist in this file!", "", "", 0);
      retval = 1;
      goto end_rtc;
    }
    else {
      fprintf(stderr, "No tables exist in file %s!\n", table->name);
      retval = 1;
      goto end_rtc;
    }	
  }

  cnline[0] = '\0';
  for (j = 0; j < table->ncol; j++) {
    if (table->selected[j] && table->type[j] > 0) {   
      char cname[1000];
      int width = 10;
      unknown = 0;
      switch(table->type[j]) {
      case TBIT:
      case TBYTE:
      case TLOGICAL:
	width = max(3, strlen(table->colnames[j]));
	break;
      case TSTRING:
	width = max(table->width[j], strlen(table->colnames[j]));
	break;
      case TSHORT:
      case TUSHORT:
	width = max(5, strlen(table->colnames[j]));
	break;
      case TLONG:
      case TULONG:
      case TINT:
      case TUINT:
	width = max(10, strlen(table->colnames[j]));
	break;
      case TFLOAT:
      case TDOUBLE:
      width = max(12, strlen(table->colnames[j]));
      break;
      case TCOMPLEX:
      case TDBLCOMPLEX:
	width = max(28, strlen(table->colnames[j]));
      break;
      default:
      unknown = 1;
      break;
      }
      table->swidth[j] = width;
      sprintf(cname, "%*s ", width, table->colnames[j]);
      if (!unknown && strlen(cname) + strlen(cnline) <= MAXBROWSERLENGTH) {
	strcat(cnline, cname);
      }
      else {
	break;  /* If the line length is too big, just ignore the rest of table columns  */
      }
    }
  }
  if (state.interactive) {
    fl_addto_browser(titles, cnline);
    /*  Read all table rows one by one  */
    fl_freeze_form(browser->form);
  }
  for (i = 1; i <= table->nrows; i++) {
    char cname[1000];
    /*  For all columns  */
    cnline[0] = '\0';
    for (j = 1; j <= table->ncol; j++) {
      j1 = j - 1;
      /*  Only read selected columns, ignore unknown types  */
      if (table->selected[j1] && table->type[j1] > 0) {   
	status = 0;
	if (table->type[j1] == TSTRING) {
	  fits_read_col(fptr, table->type[j1], j, i, 1, 1, NULL, 
			&element, NULL, &status);
	}
	else {
	  fits_read_col(fptr, table->type[j1], j, i, 1, 1, NULL, 
			element, NULL, &status);
	}
	fits_report_error(stderr, status);
	unknown = 0;
	switch(table->type[j1]) {
	case TBIT:
	case TBYTE:
	  sprintf(cname, "%*d ", table->swidth[j1], *((unsigned char *) element));
	  break;
	case TLOGICAL:
	  sprintf(cname, "%*d ", table->swidth[j1], *((int *) element));
	  break;
	case TSTRING:
	  sprintf(cname, "%*s ", table->swidth[j1], (char *) element);
	  break;
	case TSHORT:
	  sprintf(cname, "%*d ", table->swidth[j1], *((short *) element));
	  break;
	case TUSHORT:
	  sprintf(cname, "%*d ", table->swidth[j1], *((unsigned short *) element));
	  break;
	case TLONG:
	  sprintf(cname, "%*ld ", table->swidth[j1], *((long *) element));
	  break;
	case TULONG:
	  sprintf(cname, "%*ld ", table->swidth[j1], 
		   *((unsigned long *) element));
	  break;
	case TINT:
	  sprintf(cname, "%*d ", table->swidth[j1], *((int *) element));
	  break;
	case TUINT:
	  sprintf(cname, "%*d ", table->swidth[j1], *((unsigned int *) element));
	  break;
	case TFLOAT:
	  sprintf(cname, "%*e ", table->swidth[j1], *((float *) element));
	  break;
	case TDOUBLE:
	  sprintf(cname, "%*e ", table->swidth[j1], *((double *) element));
	  break;
	case TCOMPLEX:
	  sprintf(cname, "(%*e, %*e)", (table->swidth[j1] - 4) / 2, 
		   *((float *) element), (table->swidth[j1] - 4) / 2, 
		   *(((float *) element) + 1));
	  break;
	case TDBLCOMPLEX:
	  sprintf(cname, "(%*e, %*e)", (table->swidth[j1] - 4) / 2,
		   *((double *) element), (table->swidth[j1] - 4) / 2,
		   *(((double *) element) + 1));
	  break;
	default:
	  unknown = 1;
	  break;
	}
	if (!unknown && strlen(cname) + strlen(cnline) <= MAXBROWSERLENGTH) {
	  strcat(cnline, cname);
	}
	else {
	  break;  /* If the line length is too big, just ignore the rest of table columns  */
	}
      }
    }
    if (state.interactive) {
      fl_addto_browser(browser, cnline);
    }
  }
  if (state.interactive) {
    fl_set_browser_topline(browser, 1);
    fl_unfreeze_form(browser->form);
  }
  fits_report_error(stderr, status);
  
  fits_close_file(fptr, &status);
  end_rtc:
  free(element);
  return retval;
}
