/*  makeact.c  */
/*  Jure Skvarc, April 1999  */
/*  convert ACT catalog into a shortened binary form  */
/*  Only 8 fields remain  */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define DEC_OFFSET 14
#define PMRA_OFFSET 28
#define PMDEC_OFFSET 36
#define SRA_OFFSET 44
#define SDEC_OFFSET 50
#define SPMRA_OFFSET 55
#define SPMDEC_OFFSET 61
#define BMAG_OFFSET 68
#define VMAG_OFFSET 75
#define ID_OFFSET 95

typedef
struct {
  double id;  /*  TYCHO ID  */
  double ra;        /* Right ascension in decimal degrees  */
  double dec;       /* Declination in decimal degrees  */
  float pmra;      /* proper motion in ra (sec of time / year)  */
  float pmdec;     /* proper motion in dec (arcsec / year)  */
  float sra, sdec; /* standard deviation in ra and dec (mas)  */
  float spmra, spmdec;  /*  Standard deviations of proper motion mas/yr */
  float bmag;      /*  B magnitude  */
  float vmag;      /*  V magnitude  */
  int q;
}  ACT_STAR;


/*  Compare two stars on declinations  */
int
cmpstardec(const void *a1, const void *b1)

{
  ACT_STAR *a, *b;

  a = (ACT_STAR *) a1;
  b = (ACT_STAR *) b1;
  if (a->dec < b->dec) return -1;
  else if (a->dec > b->dec) return 1;
  else return 0;
}


/*  In a sorted array finds an index i where dec[i] >= d and dec[i - 1] < d  */
int
findex(ACT_STAR *stars, int nstar, double d)

{
  int first = 1;
  int last = nstar;
  int k;

  /*  0 is a special case  */
  if (d < stars[0].dec) return 0;
  /*  Check if d is out of range  */
  if (d > stars[nstar - 1].dec) return -nstar;
  do {
    k = (first + last) / 2;
    if (d < stars[k].dec) {
      if (d < stars[k - 1].dec) {
	last = k + 1;
      }
      else {
	return k;
      }
    }
    else if (d > stars[k].dec) {
      first = k;
    }
    else return k;
  } while (last - first > 1);
  return k - 1;
}


int
main(int argc,  char *argv[])

{
  FILE *fp, *outfp, *accfp;
  char line[200];
  char outname[200], *p, accname[200];
  ACT_STAR star, *stars;
  int h, m, d;
  float sec;
  double sgn;
  int n, nstars;
  int q;
  int i, k;
  char sign;
  char a[7];

  if (argc < 2) {
    fprintf(stderr, "File name missing!\n");
    exit(1);
  }

  fp = fopen(argv[1], "r");
  if (fp == NULL) {
    fprintf(stderr, "Can't open file %s for reading!\n", argv[1]);
    exit(1);
  }
  
  strcpy(outname, argv[1]);
  p = strchr(outname, '.');
  if (!p) {
    fprintf(stderr, "File name has no period!\n");
    exit(1);
  }
  *p = '\0';
  strcat(outname, "fts.bin");
  nstars = 10000;
  stars = (ACT_STAR *) malloc(sizeof(ACT_STAR) * nstars);
  if (stars == NULL) {
    fprintf(stderr, "Can't allocate space for stars!\n");
  }
  n = 0;
  while(!feof(fp)) {
    fgets(line, 200, fp);  /*  Read a line from the original file  */
    if (feof(fp)) continue;
    /*  Read and convert right ascension  */
    sscanf(line, "%d %d %f", &h, &m, &sec);
    star.ra = 15 * (h + m / 60.0 + sec / 3600.0);
    /*  Read and convert declination  */
    sscanf(line + DEC_OFFSET, "%c%d %d %f", &sign, &d, &m, &sec);
    if (sign == '-') {
      sgn = -1;
    }
    else sgn = 1;
    star.dec = sgn * (d + m / 60.0 + sec / 3600.0);

    /*  Proper motion in ra  */
    sscanf(line + PMRA_OFFSET, "%f", &star.pmra);

    /*  Proper motion in dec  */
    sscanf(line + PMDEC_OFFSET, "%f", &star.pmdec);

    /*  Error estimate of RA  */
    strncpy(a, line + SRA_OFFSET, 5);
    a[5] = '\0';
    sscanf(a, "%f", &star.sra);

    /*  Error estimate of DEC  */
    strncpy(a, line + SDEC_OFFSET, 5);
    a[5] = '\0';
    sscanf(a, "%f", &star.sdec);

    /*  Error estimate of RA motion  */
    strncpy(a, line + SPMRA_OFFSET, 6);
    a[6] = '\0';
    sscanf(a, "%f", &star.spmra);

    /*  Error estimate of DEC motion  */
    strncpy(a, line + SPMDEC_OFFSET, 6);
    a[6] = '\0';
    sscanf(a, "%f", &star.spmdec);

    /*  Blue magnitude  */
    strncpy(a, line + BMAG_OFFSET, 6);
    a[6] = '\0';
    /*  Check if there is any number  */
    for (i = 0; i < 6 && a[i] == ' '; i++);
    if (i == 6) {
      /* No data, enter -99.9  */
      star.bmag = -99.9;
    }
    else {
      /*  Read the magnitude  */
      sscanf(a, "%f", &star.bmag);
    }

    /*  Visual magnitude  */
    strncpy(a, line + VMAG_OFFSET, 6);
    a[6] = '\0';
    /*  Check if there is any number  */
    for (i = 0; i < 6 && a[i] == ' '; i++);
    if (i == 6) {
      /* No data, enter -99.9  */
      star.vmag = -99.9;
    }
    else {
      /*  Read the magnitude  */
      sscanf(a, "%f", &star.vmag);
    }

    /*  Astrometry and photometry flags  */
    q = 0;
    if (line[89] == 'X') q |= 1;  /*  Dubious astrometric reference star  */
    if (line[90] > '5') q |= 2;    /*  Probably bad astrometry  */
    if (line[91] == 'G' || line[91] == 'N') q |= 4;    /*  Known variability  */
    if (line[91] == 'U' || line[91] == 'V' || line[91] == 'W') q |= 8;   /*   Variability from Tycho data */
    star.q = q;
    /*  Tycho ID  */
    sscanf(line + ID_OFFSET, "%lf", &star.id);
    stars[n] = star;
    n++;
    if (n == nstars) {
      nstars += 10000;
      stars = (ACT_STAR *) realloc(stars, sizeof(ACT_STAR) * nstars);
      if (stars == NULL) {
	fprintf(stderr, "Can't reallocate space for stars!\n");
      }
    }
  }


  qsort(stars, n, sizeof(ACT_STAR), cmpstardec);
  /*  Open accelerator file  */
  strcpy(accname, argv[1]);
  p = strchr(accname, '.');
  *p = '\0';
  strcat(accname, ".acc");
  accfp = fopen(accname, "w");
  if (accfp == NULL) {
    fprintf(stderr, "Can't open accelerator file %s for writing!", accfp);
    exit(1);
  }
  for (i = 0; i <= 18; i++) {
    k = findex(stars, n, 10.0 * (i - 9));
    if (k >= 0) {
      fprintf(accfp, "%10f %6d %10f\n", 10.0 * (i - 9), k, stars[k].dec);
    }
    else {
      fprintf(accfp, "%10f %6d %10f\n", 10.0 * (i - 9), -k, 10.0 * (i - 9));
    }
  }
  if (fclose(accfp)) {
    fprintf(stderr, "Error in closing file %s!\n", accname);
    exit(1);
  }
  outfp = fopen(outname, "w");
  if (outfp == NULL) {
    fprintf(stderr, "Can't open file %s for writing!\n", outname);
    exit(1);
  }
  if (fwrite(stars, sizeof(ACT_STAR), n, outfp) != n) {
    fprintf(stderr, "Error in writing to file %s!\n", outname);
    exit(1);
  }
  
  if (fclose(outfp)) {
    fprintf(stderr, "Error in closing file %s!\n", outname);
    exit(1);
  }
  if (fclose(fp)) {
    fprintf(stderr, "Error in closing file %s!\n", outname);
    exit(1);
  }
}


