/*  noninteractive.c  */
/*  part of the fitsblink program  */
/*  Jure Skvarc */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <sys/types.h> 
#include <sys/time.h>
#include <unistd.h>
#include <getopt.h>
#include <time.h>

#include "functs.h"
#include "consts.h"

extern STATE state;
extern BLINK_FRAME *frame;
extern CATALOG catalog[NUMBER_OF_CATALOGS];
extern OBSERVATORY observatory;
extern MAIL mail;
extern FITS_IMAGE count_options;
extern MATCH_OPTIONS match_options;
extern TRANSROT trglob;
extern char *optarg;
extern int optopt, optind, opterr;
/*  Measure star positions and output them to a file in non-interactive mode  */

void
starcount(int argc, char *argv[])

{
  int i;
  int opt;
  int doblur = 0;

  while ((opt = getopt(argc, argv, "l:s:m:i:c:b:t:n:o:r:G:R:aABgv")) != -1) {
    switch(opt) {
    case 'l':
      /*  Name of the log file  */
      state.logfile = strdup(optarg);
      if ((state.logfp = fopen(state.logfile,"a") )== NULL) {
	fprintf(stderr, "Can not open log file!\n");
	exit(1);
      }
      break;
    case 'm':
      count_options.maxsize = strtol(optarg, NULL, 10);
      if (count_options.maxsize < 10) count_options.maxsize = 10;
      break;
    case 'b':
      count_options.border = strtol(optarg, NULL, 10);
      break;
    case 't':
      count_options.satlevel = strtol(optarg, NULL, 10);
      break;
    case 'c':
      count_options.minstar = strtol(optarg, NULL, 10);
      break;
    case 's':
      /*  Sensitivity expressed in number of standard deviations above the 
          background  */
      count_options.sg_num = strtod(optarg, NULL);
      if (count_options.sg_num < 0) count_options.sg_num = 2.0;
      break;
    case 'i':
      /*  The minimum acceptable intensity  */
      count_options.minbright = strtod(optarg, NULL);
      break;
    case 'n':
      /*  The radius of the inner circle for the aperture astrometry  */
      count_options.inner = strtod(optarg, NULL);
      break;
    case 'o':
      /*  The radius of the outer circle for the aperture astrometry  */
      count_options.outer = strtod(optarg, NULL);
      break;
    case 'r':
      /*  Coefficient about resolving close lying stars  */
      count_options.resolve = strtod(optarg, NULL);
      break;
    case 'G':
      /*  Camera gain  */
      count_options.gain = strtod(optarg, NULL);
      break;
    case 'R':
      /*  Camera read-out noise  */
      count_options.rdnoise = strtod(optarg, NULL);
      break;
    case 'a':
      /*  Aperture astrometry  */
      count_options.aperture = 1;
      break;
    case 'A':
      /*  Use ASCII output format  */
      state.starlistformat = OLD_ASCII_LIST;
      break;
    case 'B':
      /* Blur the image before counting stars  */
      doblur = 1;
      break;
    case 'g':
      /*  "Above sigma astrometry" */
      count_options.aperture = 0;
      break;
    case 'v':
      /*  Verbose mode */
      state.verbose = 1;
      break;
    case '?':
      switch(optopt) {
      default:
	fprintf(stderr, "Unknown option -%c ignored!\n", optopt);
	break;
      }
      break;
    }
  }
  for (i = optind; i < argc; i++) {
    frame[0].name = strdup(argv[i]);
    load_frame(frame, 0);
    if (doblur) blur(&frame[0].fits);
    detect_stars(frame, 0);
    star_list(frame, 0);
    remove_frame(frame, 0);
  }
}



/*  Update WCS values in the FITS file  */
int
update_wcs(BLINK_FRAME *frame, int n, TRANSROT tr)

{  
  fitsfile *fptr;
  int err = 0;
  double xinc, yinc;
  double ac, dc, xc, yc, nw, nh, fi;
  int matched;

  /*  Open file for writing  */
  fits_open_file(&fptr, frame[n].name, READWRITE, &err);
  if (err != 0) {
    if (!state.interactive) {
      fprintf(stderr, "Can't open %s for writing!\n", frame[n].name);
      return 1;
    }
  }
  /*  Calculate WCS values  */
  xinc = frame[n].map.width / frame[n].map.w;
  yinc = frame[n].map.height / frame[n].map.h;
  xc = 0.5 * frame[n].map.w;
  yc = 0.5 * frame[n].map.h;
  cal_new_wcs(tr, frame[n].map.ra, frame[n].map.dec, frame[n].map.fi, 
	      xc, yc, xinc, yinc, 
	      &ac, &dc, &nw, &nh, &fi);
  fits_update_key(fptr, TSTRING, "CTYPE1", "RA---TAN", "Inserted by fitsblink", &err);
  if (err != 0) {
    fprintf(stderr, "Disaster: could not update keyword CTYPE1 for file %s\n", frame[n].name);
    return 1;
  }
  fits_update_key_dbl(fptr, "CRPIX1", xc, 12, "Inserted by fitsblink", &err);
  if (err != 0) {
    fprintf(stderr, "Disaster: could not update keyword CRPIX1 for file %s\n", frame[n].name);
    return 1;
  }
  fits_update_key_dbl(fptr, "CRVAL1", ac, 12, "Inserted by fitsblink", &err);
  if (err != 0) {
    fprintf(stderr, "Disaster: could not update keyword CRVAL1 for file %s\n", frame[n].name);
    return 1;
  }
  fits_update_key_dbl(fptr, "CDELT1", nw, 12, "Inserted by fitsblink", &err);
  if (err != 0) {
    fprintf(stderr, "Disaster: could not update keyword CDELT1 for file %s\n", frame[n].name);
    return 1;
  }
  fits_update_key_dbl(fptr, "CROTA1", fi, 12, "Inserted by fitsblink", &err);
  if (err != 0) {
    fprintf(stderr, "Disaster: could not update keyword CROTA1 for file %s\n", frame[n].name);
    return 1;
  }
  fits_update_key(fptr, TSTRING, "CTYPE2", "DEC--TAN", "Inserted by fitsblink", &err);
  if (err != 0) {
    fprintf(stderr, "Disaster: could not update keyword CTYPE2 for file %s\n", frame[n].name);
    return 1;
  }
  fits_update_key_dbl(fptr, "CRPIX2", yc, 12, "Inserted by fitsblink", &err);
  if (err != 0) {
    fprintf(stderr, "Disaster: could not update keyword CRPIX2 for file %s\n", frame[n].name);
    return 1;
  }
  fits_update_key_dbl(fptr, "CRVAL2", dc, 12, "Inserted by fitsblink", &err);
  if (err != 0) {
    fprintf(stderr, "Disaster: could not update keyword CRVAL2 for file %s\n", frame[n].name);
    return 1;
  }
  fits_update_key_dbl(fptr, "CDELT2", nh, 12, "Inserted by fitsblink", &err);
  if (err != 0) {
    fprintf(stderr, "Disaster: could not update keyword CDELT2 for file %s\n", frame[n].name);
    return 1;
  }
  fits_update_key_dbl(fptr, "CROTA2", fi, 12, "Inserted by fitsblink", &err);
  if (err != 0) {
    fprintf(stderr, "Disaster: could not update keyword CROTA2 for file %s\n", frame[n].name);
    return 1;
  }
  matched = 1;
  fits_update_key(fptr, TLOGICAL, "MATCHED", &matched, "Inserted by fitsblink", &err);
  if (err != 0) {
    fprintf(stderr, "Disaster: could not update keyword MATCHED for file %s\n", frame[n].name);
    return 1;
  }
  fits_close_file(fptr, &err);
  if (err != 0) {
    fprintf(stderr, "Disaster: could not close file %s\n", frame[n].name);
    return 1;
  }
  return 0;
}

/*  Read a star list, field center coordinates and pixel size (or fits header), */
/*  compare it to catalog and output celestial coordinates along with any  */
/*  cross ID's                                                             */
void
make_catalog(int argc, char *argv[])

{
  int opt;
  int wcs = 0;
  char *starlist = NULL;
  double ra = 0, dec = 0, p1, p2;
  FILE *fp = NULL;
  char fn[256], ftype[7];
  char *cat = NULL, *path;
  int i;
  int xw = 0, yw = 0;
  int writewcs = 0;
  int ncatalog = 0;
  MAP_IMAGE map;
  int outputformat = FITS_STARLIST;
  
  match_options.firstbright = FIRSTBRIGHT;
  match_options.nstar = CONSTSTAR;
  match_options.min = CONSTMIN;
  match_options.mindist = MINDIST;
  match_options.poserrmax = POSERRMAX;
  match_options.magerrmax = MAGERRMAX;
  match_options.maxres = MAXRES;
  match_options.fitorder = FITORDER;
  strcpy(match_options.filter, DEFAULT_FILTER);
  map.minmag = DEFAULT_MINMAG;
  map.maxmag = DEFAULT_MAXMAG;
  map.extarea = MAP_EXTAREA; 
  p2 = p1 = -999;
  while ((opt = getopt (argc, argv, "o:b:l:n:d:D:a:w:h:m:e:i:r:f:c:x:y:vE:N:X:WAF:")) != -1) {
    switch(opt) {
    case 'l':
      /*  Name of the log file  */
      state.logfile = strdup(optarg);
      if ((state.logfp = fopen(state.logfile,"a") )== NULL) {
	fprintf(stderr, "Can not open log file!\n");
	exit(1);
      }
      break;
    case 'o':
      /*  Order of fitting  */
      match_options.fitorder = strtol(optarg, NULL, 10);
      if (match_options.fitorder < 1) match_options.fitorder = 1;
      if (match_options.fitorder > 5) match_options.fitorder = 5;
      break;
    case 'b':
      match_options.firstbright = strtol(optarg, NULL, 10);
      if (match_options.firstbright < MINFIRSTBRIGHT) match_options.firstbright = MINFIRSTBRIGHT;
      break;
    case 'a':
      /* Get right ascension  */
      ra = strtod(optarg, NULL);
      wcs++;
      break;
    case 'd':
      /* Get declination */
      dec = strtod(optarg, NULL);
      wcs++;
      break;
    case 'w':
      /* Get pixel width */
      p1 = strtod(optarg, NULL);
      wcs++;
      break;
    case 'h':
      /* Get pixel height */
      p2 = strtod(optarg, NULL);
      wcs++;
      break;
    case 'x':
      /*  Number of pixels in x direction  */
      xw = strtol(optarg, NULL, 10);
      wcs++;
      break;
    case 'y':
      /*  Number of pixels in y direction  */
      yw = strtol(optarg, NULL, 10);
      wcs++;
      break;
    case 'f':
      /*  Name of the star list file  */
      starlist = strdup(optarg);
    case 'n':
      match_options.nstar = strtol(optarg, NULL, 10);
      break;
    case 'm':
      match_options.min = strtod(optarg, NULL);
      break;
    case 'D':
      match_options.mindist = strtod(optarg, NULL);
      break;
    case 'e':
      match_options.poserrmax = strtod(optarg, NULL);
      break;
    case 'i':
      match_options.magerrmax = strtod(optarg, NULL);
      break;
    case 'r':
      match_options.maxres = strtod(optarg, NULL);
      break;
    case 'F':
      strncpy(match_options.filter, optarg, 1);
      break;
    case 'c':
      /*  Extract catalog name and path  */
      cat = optarg;
      path = strtok(cat, ":");
      path = strtok(NULL, ":");
      /*  Check which catalog we have  */
      /*===============================*/
      if (strcmp(cat, "usno") == 0) {
	ncatalog++;
	catalog[CAT_USNO].use = 1;
	catalog[CAT_USNO].path = strdup(path);
	catalog[CAT_USNO].lpath = strlen(path);
      } 
      else if (strcmp(cat, "gscn") == 0) {
	ncatalog++;
	catalog[CAT_GUIDE].use = 1;
	catalog[CAT_GUIDE].path = strdup(path);
	catalog[CAT_GUIDE].lpath = strlen(path);
      } 
      else if (strcmp(cat, "gscs") == 0) {
	ncatalog++;
	catalog[CAT_GUIDE_SOUTH].use = 1;
	catalog[CAT_GUIDE_SOUTH].path = strdup(path);
	catalog[CAT_GUIDE_SOUTH].lpath = strlen(path);
      } 
      else if (strcmp(cat, "act") == 0) {
	  ncatalog++;
	  catalog[CAT_ACT].use = 1;
	  catalog[CAT_ACT].path = strdup(path);
	  catalog[CAT_ACT].lpath = strlen(path);
      } 
      else if (strcmp(cat, "tycho2") == 0) {
	  ncatalog++;
	  catalog[CAT_TYCHO2].use = 1;
	  catalog[CAT_TYCHO2].path = strdup(path);
	  catalog[CAT_TYCHO2].lpath = strlen(path);
      } 
      else {
	fprintf(stderr, "Unknown catalog\n");
	fprintf(stderr, "Try one of these:\n");
	fprintf(stderr, "  usno -- USNO\n");
	fprintf(stderr, "  gscn -- GSC 1.1 from 90 to -7.5 declination\n");
	fprintf(stderr, "  gscs -- GSC 1.1 from -7.5 to -90 declination\n");
	fprintf(stderr, "  act -- ACT\n"); 
	fprintf(stderr, "  tycho2 -- TYCHO-2\n"); 
      }
      break;
    case 'E':
      map.extarea = strtod(optarg, NULL);
      break;
    case 'X':
      map.maxmag = strtod(optarg, NULL);
      break;
    case 'N':
      map.minmag = strtod(optarg, NULL);
      break;
    case 'v':
      state.verbose = 1;
      break;
    case 'W':
      writewcs = 1;
      break;
    case 'A':
      /*  Output list in ASCII  */
      outputformat = OLD_ASCII_LIST;
      break;
    case '?':
      switch(optopt) {
      default:
	fprintf(stderr, "Unknown option -%c ignored!\n", optopt);
	break;
      }
      break;
    }
  }

  if (starlist == 0) {
    fprintf(stderr, "No file name!  Exiting...\n");
    exit(1);
  }
  if ((fp = fopen(starlist, "r") )== NULL) {
    fprintf(stderr, "Can not open starlist file: %s!\n", starlist);
    exit(1);
  }
  /*  Check if file is FITS  */
  for (i = 0; i < 6; i++) {
    ftype[i] = fgetc(fp);
  }
  ftype[6] = '\0';
  if (strcmp(ftype, "SIMPLE") == 0) {
    int status = 0;
    fitsfile *fptr;
    char comment[FLEN_CARD];

    fclose(fp);
    state.starlistformat = FITS_STARLIST;
    /*  Open the file  */
    fits_open_file(&fptr, starlist, READONLY, &status);
    if (status) {
      fprintf(stderr, "Can not open file %s for reading!\n", starlist);
      exit(1);
    }
    /*  Move to the table extension  */
    fits_movnam_hdu(fptr, BINARY_TBL, "Objects", 0, &status);
    frame[0].name = (char *) myalloc(sizeof(char) * 1024, "make_catalog",
				     "frame[0].name");
    fits_read_key_str(fptr, "IMGNAME", frame[0].name, comment, &status);
    fits_close_file(fptr, &status);
    fits_report_error(stderr, status);
    if (status != 0) {
      fprintf(stderr, "Error in reading input file.  Exiting...\n");
      exit(1);
    }
  }
  else {
    state.starlistformat = OLD_ASCII_LIST;
    /*  In the first file line we look for the image file name  */
    fscanf(fp, "file name: %s", fn);
    frame[0].name = strdup(fn);
    fclose(fp);
  }
  /*  Load the image  */
  if (load_frame(frame, 0)) {
    fprintf(stderr, "Can't load %s\n", frame[0].name);
    exit(1);
  }
  frame[0].map.extarea = map.extarea;
  frame[0].map.maxmag = map.maxmag;
  frame[0].map.minmag = map.minmag;
  if (wcs != 6) {
    if (!frame[0].fitsfile.wcs) {
      fprintf(stderr, "No image coordinates!\n");
      exit(1);
    }
  }
  else {
    if (!frame[0].fitsfile.wcs) {
      frame[0].fitsfile.nax[0] = xw;
      frame[0].fitsfile.nax[1] = yw;
    }
    /*  Write or override image coordinates and pixel dimensions  */
    frame[0].ra = frame[0].fitsfile.crval1 = ra;
    frame[0].dec = frame[0].fitsfile.crval2 = dec;
    frame[0].fitsfile.crpix1 = 0.5 * frame[0].fitsfile.nax[0];
    frame[0].fitsfile.crpix2 = 0.5 * frame[0].fitsfile.nax[1];
    frame[0].fitsfile.crot = 0;
    strcpy(frame[0].fitsfile.ctype, "-TAN");
    frame[0].fitsfile.wcs = 1;
  }
  if (p1 != -999) {
    frame[0].xsize = p1; 
    frame[0].fitsfile.cdelt1 = p1 / 3600.0;
  }
  if (p2 != -999) {
    frame[0].ysize = p2; 
    frame[0].fitsfile.cdelt2 = p2 / 3600.0;
  }

  /*  Now read the complete file with objects */
  load_star_list(frame, 0);
  
  if (ncatalog == 0) {
    fprintf(stderr, "No catalog is specified!\n");
    exit(1);
  }
  frame[0].match = match_options;
  /*  Match a list with a star catalog  */
  if (match_stars(frame, 0)) {
    fprintf(stderr, "Matching failed!\n");
    exit(1);
  }
  state.starlistformat = outputformat;
  /*  Output star list  */
  star_list(frame, 0);
  if (writewcs) {
    update_wcs(frame, 0, trglob);
  }
  remove_frame(frame, 0);
}




