/*  aperture.c   */
/*  Part of the fitsblink program */
/*  Jure Skvarc, November 1998  */
/*  Given coordinates of the center point and two radii, it calculates */
/*  star or comet intensity inside the inner circle and subtracts the  */
/*  median intensity between the circles  */
/*  It assumes that the inner radius is smaller than the outer radius  */

#include <stdio.h>
#define __USE_BSD 
#include <math.h>
#include <stdlib.h>
#include <malloc.h>
#include <forms.h>
#include "../functs.h"

extern POINT stack[];
extern STATE state;
extern GAUSSF gaussf;
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif
/*  Find out how many vertices of the pixel are inside the given circle */
/*  and return flags, which tell which points are inside the circle     */
/*  index       coordinates    */
/*  0           xt - 0.5, yt - 0.5  */         
/*  1           xt + 0.5, yt - 0.5  */         
/*  2           xt + 0.5, yt + 0.5  */         
/*  3           xt - 0.5, yt + 0.5  */
int
inside(double x, double y, double r, double xt, double yt, double *inside)

{
  double xm, xv, ym, yv;
  int i, k = 0;

  r *= r;
  xt -= x;
  yt -= y;
  xm = (xt - 0.5);
  xm *= xm;
  xv = (xt + 0.5);
  xv *= xv;
  ym = (yt - 0.5);
  ym *= ym;
  yv = (yt + 0.5);
  yv *= yv;
  inside[0] = r - xm - ym;
  inside[1] = r - xv - ym;
  inside[2] = r - xv - yv;
  inside[3] = r - xm - yv;
  for (i = 0; i <= 3; i++) 
    if (inside[i] > 0) k++;
  return k;
}

/*  Find the pixel with the highest value inside the circle with */
/*  the radius r and center coordinates x and y.  Return the  */
/*  pixel coordinates in xf and yf    */
void
find_peak(FITS_IMAGE *slika, double x, double y, double r, int *xf, int *yf)

{
#if 0
  int i, imin, imax, i2;
  int j, jmin, jmax;
  int vmax = -10000000, v;
  double dist[4];

  r += 0.5;
  imin = x - r;
  imax = x + r;
  *xf = x;
  *yf = y;
  for (i = imin; i <= imax; i++) {
    i2 = i - x;
    i2 *= i2;
    jmax = (int) (sqrt(r * r - i2) + 0.5);
    jmin = y - jmax;
    jmax += y;
    for (j = jmin; j <= jmax; j++) {
      if (inside(x, y, r, (double) i, (double) j, dist) > 0) {
	v = video_point(slika, i, j);
	if (v > vmax) {
	  vmax = v;
	  *xf = i;
	  *yf = j;
	}
      }
    }
  }
#endif
  int i, imax;
  int j, jmax, x0, y0;
  int vmax = -10000000, v;
  double ins[4];

  r += 0.5;
  imax = x + r;
  imax = x;
  jmax = y;
  do {
    x0 = imax;
    y0 = jmax;
    for (i = x0 - 2; i <= x0 + 2; i++) {
      for (j = y0 - 2; j <= y0 + 2; j++) {
	if (inside(x, y, r, (double) i, (double) j, ins) > 0) {
	  v = video_point(slika, i, j);
	  if (v > vmax) {
	    vmax = v;
	    imax = i;
	    jmax = j;
	  }
	}
      }
    }
  } while (!(imax == x0 && jmax == y0));
  *xf = x0;
  *yf = y0;
}


int
vrh3(FITS_IMAGE *sl, int x, int y, int v, int back, float sg) 

{
  int v1;

  v1 = (long) video_point(sl, x, y) - back;
  return v1 > sl->sg_num * sg && (v - back) + 0.2 * v1 / sg > v1;
}

/*  Subroutine which visits all points which are above the given threshold */
/*  and within a certain distance from the starting point  */
/*  The algorithm uses stack  */
int
aperture_visit(FITS_IMAGE *slika, double x0, double y0, float r, int thresh, int back, float sg, OBJEKT *obj, double *x2, double *y2, double *sxt, double *syt)
{
  int sp = 0;
  int v;
  SHADOW senca;
  int side = 2 * r + 4;
  double ins[4];
  int x1, y1;
  int xmin, xmax, ymin, ymax;
  
  xmin = x0 - r - 1;
  xmax = xmin + side - 1;
  ymin = y0 - r - 1;
  ymax = ymin + side - 1;
  reserve_shadow(&senca, side, side, (int) (x0 - r - 1), (int) (y0 - r - 1));
  clear_shadow(&senca);
  /*  Postavim zacetne podatke o objektu   */
  obj->area = 0;
  obj->grey = 0;
  obj->sgrey = 0.0;
  obj->x1 = obj->y1 = *x2 = *y2 = 0;
  obj->xt = obj->yt = *sxt = *syt = 0;
  stack[sp].x = x0;
  stack[sp++].y = y0;
  do {
    float x, y;
    sp--;
    x = stack[sp].x;
    y = stack[sp].y;
    if (!test(&senca, x, y)) {
      /*  Oznacim, da je tocka obiskana  */
      set(&senca, x, y);
      v = video_point(slika, x, y);
      obj->area++;
      obj->grey += v;
      obj->xt += v * x;
      obj->yt += v * y;
      *sxt += v * x * x;
      *syt += v * y * y;
      obj->x1 += x;
      *x2 += x * x;
      obj->y1 += y;
      *y2 += y * y;
      /*  Pogledam, ali lahko stopim na okoliske tocke */
      y1 = y - 1;
      if (y > ymin && video_point(slika, x, y1) > thresh && 
	  vrh3(slika, x, y1, v, back, sg) &&
	  inside(x0, y0, r, (double) x, (double) y1, ins) > 0) {
	stack[sp].x = x;
	stack[sp++].y = y1;
	if (sp == STACKMAX) {
	  fprintf(stderr, "Stack overflow in line %d of file %s\n", 
		  __LINE__, __FILE__);
	  sp--;
	}
      }
      y1 = y + 1;
      if (y < ymax && video_point(slika, x, y1) > thresh &&
	  vrh3(slika, x, y1, v, back, sg) &&
	  inside(x0, y0, r, (double) x, (double) y1, ins) > 0) {
	stack[sp].x = x;
	stack[sp++].y = y1;
	if (sp == STACKMAX) {
	  fprintf(stderr, "Stack overflow in line %d of file %s\n", 
		  __LINE__, __FILE__);
	  sp--;
	}
      }
      x1 = x - 1;
      if (x > xmin && video_point(slika, x1, y) > thresh &&
	  vrh3(slika, x1, y, v, back, sg) &&
	  inside(x0, y0, r, (double) x1, (double) y, ins) > 0) {
	stack[sp].x = x1;
	stack[sp++].y = y;
	if (sp == STACKMAX) {
	  fprintf(stderr, "Stack overflow in line %d of file %s\n", 
		  __LINE__, __FILE__);
	  sp--;
	}
      }
      x1 = x + 1;
      if (x < xmax && video_point(slika, x1, y) > thresh &&
	  vrh3(slika, x1, y, v, back, sg) &&
	  inside(x0, y0, r, (double) x1, (double) y, ins) > 0) {
	stack[sp].x = x1;
	stack[sp++].y = y;
	if (sp == STACKMAX) {
	  fprintf(stderr, "Stack overflow in line %d of file %s\n", 
		  __LINE__, __FILE__);
	  sp--;
	}
      }
    }
  } while(sp > 0);
  free_shadow(&senca);
  return 0;
}





/*  Make an aperture photometry of the object with the brightest pixel  */
/*  at coordinate x0, y0.  Add values inside a circle with a radius inner */
/*  Subtract background calculated as a median value of the pixels between */
/*  circles with radii inner and outer */
/*double
aperture(FITS_IMAGE *sl, double x0, double y0, double inner, double outer, 
double *sx, double *sy, float *swx, float *swy, float *sym, float *sigma_grey, int fpeak) */

double
aperture(FITS_IMAGE *sl, OBJEKT *obj1, double inner, double outer, int fpeak)

{
  double v,  bgval, ball;
  double bx = 0.0, by = 0.0;
  int x, y;
  double x2, y2, sxt, syt;
  double sg;
  int inarea, outarea;
  double i, j, i2;
  int k, kmax;
  int rmax, imin, imax, jmin, jmax;
  double dist[4];
  int size;
  static int bsize = 0;
  static PIXEL *backg = NULL;
  double o2 = outer * outer;
  double sv;
  int above_thresh;
  double thresh;
  OBJEKT obj;
  double x0, y0, sx, sy;
  float swx, swy, sym, sigma_grey;
  
  x0 = obj1->xf;
  y0 = obj1->yf;
  
  /*  radius of the circle to be examined  */
  rmax = outer + 0.5;
  /*  areas of the inner circle and of the outer minus inner circle  */
  inarea = outarea = 0;
  /*  range of x coordinates to be examined  */
  imin = x0 - rmax;
  imax = x0 + rmax;
  /*  Calculate the size needed for the storage of the background pixels */
  size = M_PI * o2;
  if (size > bsize) {
    if (backg) {
      backg = (PIXEL *) realloc(backg, sizeof(PIXEL) * size);
      if (!backg) {
	fprintf(stderr, "Out of memory, function aperture, backg.\n");
	exit(1);
      }
    }
    else {
      backg = (PIXEL *) myalloc(sizeof(PIXEL) * size, "aperture", "backg");
    }
    bsize = size;
  }
  for (i = imin; i <= imax; i++) {
    i2 = i - x0;
    i2 *= i2;
    jmax = (int) (sqrt(o2 - i2) + 0.5);
    jmin = y0 - jmax;
    jmax += y0;
    for (j = jmin; j <= jmax; j++) {
      if (inside(x0, y0, inner, i, j, dist) > 0) {
	v = video_point(sl, (int) i, (int) j);
	/* this pixel is inside the inner circle with at least one vertex */
	inarea++;
      }
	/* this pixel is inside the outer circle with at least one vertex */
      else if (inside(x0, y0, outer, i, j, dist) > 0) {
	v = video_point(sl, (int) i, (int) j);
	/*  This pixel is between the inner and outer circle  */
	backg[outarea] = v;
	outarea++;
	if (outarea == bsize) {
	  bsize += 100;	
	  backg = (PIXEL *) realloc(backg, sizeof(PIXEL) * bsize);
	  if (!backg) {
	    fprintf(stderr, "Out of memory, function aperture, backg.\n");
	    exit(1);
	  }
	}
      }
    }
  }
  /*  Calculate the median of the background pixels  */
  qsort(backg, outarea, sizeof(PIXEL), cmppixel);
  /*  Calculate the sigma  */
  sg = 0;
  i2 = 0;
  bgval = 0;
  kmax = outarea * 0.7;
  for (k = 0.3 * outarea; k < kmax; k++) {
    bgval += backg[k];
    sg += backg[k] * backg[k];
    i2++;
  }
  bgval /= i2;
  sg = sqrt(sg / i2 - bgval * bgval);
  thresh = bgval + sl->sg_num * sg;
  if (fpeak) {
    find_peak(sl, (double) x0, (double) y0, inner, &x, &y);
    x0 = x;
    y0 = y;
  }
  /*  Calculate the intensity and centroid of the pixel larger than background  */
  /*  for sg_num * sigma  */
  aperture_visit(sl, (double) x0, (double) y0, inner, thresh, bgval, sg, &obj, &x2, &y2, &sxt, &syt);
  above_thresh = obj.area;
  sv = obj.grey;
  obj1->xt = obj.xt;
  obj1->yt = obj.yt;
  bx = obj.x1;
  by = obj.y1;
  if (above_thresh > 0) {
    ball =  bgval * above_thresh;
    /*  Subtract the background from the signal  */
    sv -= ball;
    /*  Estimate the error, formula from the "Realistic CCD model by Merline & Howell"  */
    obj1->sgrey = sqrt(sv * sl->gain + above_thresh * (1.0 + (double) above_thresh / i2) * 
      (sl->rdnoise * sl->rdnoise  + sl->gain * sl->gain / 12.0) + 
      fsqr((double) above_thresh / i2 * sg)) / sl->gain;

    obj1->xt = (obj1->xt - bx * bgval) / sv;
    obj1->yt = (obj1->yt - by * bgval) / sv;
    obj1->sxt = sqrt((sxt - x2 * bgval) / sv - fsqr(obj1->xt));
    obj1->syt = sqrt((syt - y2 * bgval) / sv - fsqr(obj1->yt));
    obj1->area = above_thresh;
    /*  Find out the symmetry of the object  */
    x = (int) (obj1->xt + 0.5);
    y = (int) (obj1->yt + 0.5);
    v = 0;
    for (i = 1; i <= inner; i++) {
      v += video_point(sl, x , y + i) + video_point(sl, x, y - i) -
	(video_point(sl, x + i, y) + video_point(sl, x - i, y)) + 
	video_point(sl, x + 1 , y + i) + video_point(sl, x + 1, y - i) -
	(video_point(sl, x + i, y + 1) + video_point(sl, x - i, y + 1)) + 
	video_point(sl, x - 1 , y + i) + video_point(sl, x - 1, y - i) -
	(video_point(sl, x + i, y - 1) + video_point(sl, x + i, y - 1));
    }
    obj1->sym = v;
  }
  else {
    sv = video_point(sl, (int) x0, (int) y0) - bgval;
    obj1->x1 = x0;
    obj1->y1 = y0;
    obj1->sxt = 1.0 / sqrt(12.0);
    obj1->syt = obj1->sxt;
    obj1->sym = 0;
    obj1->area = 1;
  }
  return sv;
}



