/*  trail.c (part of fitsblink) */
/*  Jure Skvarc, February 2001  */
/*  Routines for measurement of asteroid trails  */
#include <math.h>

#define SQRT_PI_2 (M_SQRT2 / M_2_SQRTPI)
/*  A function, which represents a trail with given parameters:  */
/*    x, y - a point where we evaluate the function  */
/*    sx, sy  -  widths of the PSF in x and y direction  */
/*    vx, vy -  asteroid velocities in pixels/s  */
/*    x0, y0 -  assumed starting point of the trail  */
/*    t      -  exposure length in seconds  */
double
trail(double x, double y, double sx, double sy, double vx, double vy,
      double x0, double y0, double t)

{
  double g1, g2;
  double e1, e;
  double erf0, erft;

  g1 = sqrt(1.0 / sqrt(sy * sy * vx * vx + sx * sx * vy * vy));
  g2 = M_SQRT2 * sx * sy * g1;
  e1 = vy * (x - x0) + vx * (y0 - y);
  e = exp(-(e1 * e1) / 2.0 * g1 * g1);
  erf0 = erf((sy * sy * vx * (x0 - x) + sx * sx * vy * (y0 - y)) / g2);
  erft = erf((sy * sy * vx * (x0 - x + t * vx) + sx * sx * vy * (y0 - y + t * vy)) / g2);
  return -(e * SQRT_PI_2 * sx * sy * (erf0 - erft));
}


/*  Return partial derivatives of the trail function  */
/*    x, y - a point where we evaluate the function  */
/*    sx, sy  -  widths of the PSF in x and y direction  */
/*    vx, vy -  asteroid velocities in pixels/s  */
/*    x0, y0 -  assumed starting point of the trail  */
/*    t      -  exposure length in seconds  */

double
der_trail(double x, double y, double sx, double sy, double vx, double vy,
      double x0, double y0, double t)

{
  double g0, g1, g2;
  double e1, exp1, e2, exp2, e3, exp3;
  double sx2, sy2;
  double erf0, erft;
  double vx2, vy2m sx2, xy2, sxsy;
  double dx, dy, dxt, dyt;
  double derf;

  sx2 = sx * sx;
  sy2 = sy * sy;
  vx2 = vx * vx;
  vy2 = vy * vy;
  sxsy = sx * sy;
  g0 = sy2 * vx2 + sx2 * vy2;
  g12 = sqrt(g0);
  g32 = g0 * g12;
  g52 = g32 * g0;
  g2 = M_SQRT2 * sxsy * g12;
  dx = x0 - x;
  dxt = t * vx + dx;
  dy = y0 - y;
  dyt = t * vy + dy;
    
  e1 = vy * -dx + vx * dy;
  exp1 = exp(-(e1 * e1) / (2.0 * g0));
  e2 = sy2 * dx + sx2 * vy * dy;
  exp2 = exp(-e2 * e2 / (2.0 * sx2 * sy2 * g0));
  e3 = sy2 * vx * dxt + sx2 * vy * dyt;
  exp3 = exp(-e3 * e3 / (2.0 * sx2 * sy2 * g0));
  erf0 = erf((sy2 * vx * dx + sx2 * vy * dy) / g2);
  erft = erf((sy2 * vx * dxt + sx2 * vy * dyt) / g2);
  derf = erf0 - erft;

  trsx = 1.0 / g52 * (exp1 * SQRT_PI_2 * sy * 
		      (-1.0 / sx * (sy * vx * g12 / SQRT_PI_2 * 
				    (exp3 * (sy2 * vx2 * dxt + sx2 * vy * 
					     (t * vx * vy + 2.0 * vy * dx + vx * -dy)) +
			      exp2 * (sy2 * vx2 * -dx + sx2 * vy * (2.0 * vy * -dx + vx * dy)))) +
		      derf * (g0 * (sx2 * vy2 - g0) - 
			      sx2 * vy2 * pow(vy * -dx + vx * dy, 2.0))));

  trsy = 1.0 / g52 * (exp1 * SQRT_PI_2 * sx * 
		      (-1.0 / sy * (sx * vy * g12 / SQRT_PI_2 * 
				    (exp3 * (sx2 * vy2 * dyt + sy2 * vx * 
					     (t * vx * vy + 2.0 * vx * dy + vy * -dx)) +
			      exp2 * (sx2 * vy2 * -dy + sy2 * vx * (2.0 * vx * -dy + vy * dx)))) +
		      derf * (g0 * (sy2 * vx2 - g0) - 
			      sy2 * vx2 * pow(vy * -dx + vx * dy, 2.0))));

  trvx = 1.0 / g32 * (exp1 * SQRT_PI_2 * sxsy * 
		      (-(sy / SQRT_PI_2 * 
			 (exp2 * sx2 * vy  * (vy * dx + vx * -dy) -
			  exp3 * (sy2 * t * vx2 * vx + 
				  sx2 * vy * (t * vx * vy + vy * dx + vx * -dy))) / (sx * g12) +
			 sy2 * vx * derf + 
			 1.0 / g0 * (vy * (vy * dx + vx * -dy) * 
				     (sy2 * vx * -dx + sx2 * vy * -dy) * derf)));

  trvx = 1.0 / g32 * (exp1 * SQRT_PI_2 * sxsy * 
		      (-(sy / SQRT_PI_2 * 
			 (exp2 * sx2 * vy  * (vy * dx + vx * -dy) -
			  exp3 * (sy2 * t * vx2 * vx + 
				  sx2 * vy * (t * vx * vy + vy * dx + vx * -dy))) / (sx * g12) +
			 sy2 * vx * derf + 
			 1.0 / g0 * (vy * (vy * dx + vx * -dy) * 
				     (sy2 * vx * -dx + sx2 * vy * -dy) * derf))));

  trvy = 1.0 / g32 * (exp1 * SQRT_PI_2 * sxsy * 
		      (-(sx / SQRT_PI_2 * 
			 (exp2 * sy2 * vx  * (vy * dx + vx * -dy) -
			  exp3 * (sx2 * t * vy2 * vy + 
				  sy2 * vx * (t * vx * vy + vy * -dx + vx * dy))) / (sy * g12) +
			 sx2 * vy * derf - 
			 1.0 / g0 * (vx * (vy * dx + vx * -dy) * 
				     (sy2 * vx * -dx + sx2 * vy * -dy) * derf))));

  trx0 = 1.0 / g32 * (exp1 * SQRT_PI_2 * sxsy * 
		      (-1.0 / sx * ((exp2 - exp3) / SQRT_PI_2 * sy * vx * g12) -
		       vy * (vy * -dx + vx * -dy) * derf));
  try0 = 1.0 / g32 * (exp1 * SQRT_PI_2 * sxsy * 
		      (-1.0 / sy * ((exp2 - exp3) / SQRT_PI_2 * sx * vy * g12) +
		       vx * (vy * -dx + vx * -dy) * derf));

  f = -(exp1 * SQRT_PI_2 * sxsy * derf);
}
